/*******************************************************************************
 * Copyright (c) 2010- Association for Decentralized Information Management in
 * Industry THTH ry.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors:
 *    VTT Technical Research Centre of Finland - initial API and implementation
 *******************************************************************************/
package org.simantics.databoard.datasource;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Map;
import java.util.concurrent.CopyOnWriteArrayList;

import org.simantics.databoard.Bindings;
import org.simantics.databoard.Datasources;
import org.simantics.databoard.binding.Binding;
import org.simantics.databoard.binding.mutable.MutableVariant;
import org.simantics.databoard.binding.mutable.MutableVariantBinding;
import org.simantics.databoard.binding.mutable.Variant;
import org.simantics.databoard.type.Datatype;

/**
 * This is an implementation that composes nodes and forms a 
 * datasource. 
 *
 * @author Toni Kalajainen <toni.kalajainen@vtt.fi>
 */
public abstract class StepwiseDatasource implements IStepwiseDatasource {

	/** Listeners */ 
	CopyOnWriteArrayList<DatasourceListener> listeners = new CopyOnWriteArrayList<DatasourceListener>(); 
	
	/** ID Binding */
	Binding idBinding;
	
	/** Nodes */
	Map<Variant, Node> nodes = new HashMap<Variant, Node>();
	
	RootNode root;
	
	/**
	 * Create new stepwise data source
	 * 
	 */
	public StepwiseDatasource() {
		this.idBinding = Bindings.MUTABLE_VARIANT;
		root = new RootNode();
		nodes.put(root.getId(), root);
	}
	
	public void step() throws DatasourceException {
		for (DatasourceListener l : listeners)
			l.onPreStep();
		doStep();
		for (DatasourceListener l : listeners)
			l.onPostStep();
	}	
	
	protected abstract void doStep() throws DatasourceException;
	
	public void addNode(Node node) {
		nodes.put(node.getId(), node);
	}
	
	public void addRoot(Node node) {
		addNode(node);
		if ( !root.children.contains(node.getId()))
			root.children.add(node.getId());
	}
	
	@Override
	public boolean hasNode(Variant nodeId) {
		return nodes.containsKey(nodeId);
	}
	
	@Override
	public void getAll(Collection<Variant> list) {
		list.addAll( nodes.keySet() );
	}

	Node getNode(Variant nodeId) throws DatasourceException {
		if (nodeId.equals(root.getId())) return root;		
		Node node = nodes.get( nodeId );
		if (node==null) throw new DatasourceException(nodeId+" does not exist");
		return node;
	}
	
	@Override
	public void getChildren(Variant nodeId, Collection<Variant> children)
			throws DatasourceException {
		Node n = getNode(nodeId);
		n.getChildren(children);
	}

	@Override
	public boolean getValue(Variant nodeId, MutableVariant value) throws DatasourceException {
		Node node = getNode(nodeId);
		return node.getValue(value);
	}

	@Override
	public boolean setValue(Variant nodeId, Variant value)
			throws DatasourceException {
		Node node = getNode(nodeId);
		return node.setValue(value);
	}
	
	@Override
	public Datatype getNodeType(Variant nodeId) throws DatasourceException {
		Node node = getNode(nodeId);
		MutableVariant value = new MutableVariant();
		if (!node.getValue(value)) return null;
		return value.type();
	}

	@Override
	public String getLabel(Variant nodeId, String locale) throws DatasourceException {
		Node node = getNode(nodeId);
		return node.getLabel(locale);
	}

	@Override
	public void getLabels(Variant nodeId, Map<String, String> labels) throws DatasourceException {
		Node node = getNode(nodeId);
		node.getLabels(labels);
	}
	
	@Override
	public void addListener(DatasourceListener listener) {
		listeners.add( listener );		
	}

	@Override
	public void removeListener(DatasourceListener listener) {
		listeners.remove( listener );
	}	

	static class RootNode implements Node {

		ArrayList<Variant> children = new ArrayList<Variant>();
		
		@Override
		public Variant getId() {
			return Datasources.ROOT_ID;
		}

		@Override
		public void getChildren(Collection<Variant> children) {
			children.addAll( this.children );			
		}

		@Override
		public boolean getValue(MutableVariant value) {
			return false;
		}

		@Override
		public String getLabel(String locale) {
			// TODO Auto-generated method stub
			return null;
		}

		@Override
		public void getLabels(Map<String, String> locales) {
			// TODO Auto-generated method stub
			
		}

		@Override
		public boolean setValue(Variant value) {
			return false;
		}
		
	}
	
	public interface Node {
		Variant getId();
		void getChildren(Collection<Variant> children);
		boolean getValue(MutableVariant value);
		boolean setValue(Variant value) throws DatasourceException;
		String getLabel(String locale);
		void getLabels(Map<String, String> locales);
	}

}

